# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

from datetime import datetime, timedelta
from impacket.dcerpc.v5.ndr import NULL
from impacket.dcerpc.v5 import wkst, srvs, samr
from impacket.dcerpc.v5.samr import DCERPCSessionError
from impacket.dcerpc.v5.rpcrt import DCERPCException
from impacket.dcerpc.v5.dcom.wmi import WBEM_FLAG_FORWARD_ONLY
from bs4 import BeautifulSoup
from ldap3.utils.conv import escape_filter_chars
from ldap3.protocol.microsoft import security_descriptor_control
from ldap3.protocol.formatters.formatters import *
from impacket.ldap.ldaptypes import ACE, ACCESS_ALLOWED_OBJECT_ACE, ACCESS_MASK
from impacket.ldap.ldaptypes import LDAP_SID, SR_SECURITY_DESCRIPTOR, LDAP_SERVER_SD_FLAGS

from pywerview.requester import LDAPRPCRequester
import pywerview.objects.adobjects as adobj
import pywerview.objects.rpcobjects as rpcobj
import pywerview.formatters as fmt

class NetRequester(LDAPRPCRequester):
    def _resolve_sid(self, sid, sid_mapping, attribute=['distinguishedname']):
        try:
            resolved_sid = sid_mapping[sid]
        except KeyError:
            self._logger.warning('SID ({}) is not a well known or already resolved SID'.format(sid))
            try:
                resolved_sid = self.get_adobject(queried_sid=sid, queried_domain=self._queried_domain,
                                                 attributes=attribute)[0].distinguishedname
                self._logger.debug('SID ({0}) is ({1})'.format(sid, resolved_sid))
            except IndexError:
                self._logger.warning('We did not manage to resolve this SID ({}) against the DC'.format(sid))
                resolved_sid = sid
        finally:
            sid_mapping[sid] = resolved_sid

        return resolved_sid

    @LDAPRPCRequester._ldap_connection_init
    def get_netpki(self, queried_domain=str(), queried_ca_name=str(), resolve_sids=False, full_data=False):

        # This function is mostly based on the dumpADCS() one within impacket's ntlmrelayx
        # credit goes to the multiple contributors!
        if full_data:
            attributes=list()
        else:
            attributes = ["certificateTemplates", "displayName", "dNSHostName", "name",
                          "msPKI-Enrollment-Servers", "nTSecurityDescriptor"]

        if queried_ca_name:
            ldap_filter = '(&(objectClass=pKIEnrollmentService)(displayname={}))'.format(queried_ca_name)
        else:
            ldap_filter = '(objectClass=pKIEnrollmentService)'

        controls = security_descriptor_control(criticality=True, sdflags=LDAP_SERVER_SD_FLAGS.DACL_SECURITY_INFORMATION.value)
        base_dn = self._base_dn
        self._base_dn = self._server_info.other['configurationNamingContext'][0]
        objectpki_raw = self._ldap_search(ldap_filter, adobj.PKIEnrollmentService, attributes=attributes, controls=controls)
        self._base_dn = base_dn

        sid_mapping = adobj.ADObject._well_known_sids.copy()
        objectpki = list()
        for pki in objectpki_raw:
            sd = SR_SECURITY_DESCRIPTOR()
            sd.fromString(pki.ntsecuritydescriptor)

            enrollment_principals = list()
            # AccessAllowedObject = 0x05
            for ace in (a for a in sd["Dacl"]["Data"] if a["AceType"] == 0x05):
                sid = format_sid(ace["Ace"]["Sid"].getData())
                if ace["Ace"]["Flags"] == 2:
                    uuid = format_uuid_le(ace["Ace"]["InheritedObjectType"])[1:-1].lower()
                    self._logger.log(self._logger.ULTRA, "UUID found in InheritedObjectType: {}".format(uuid))
                elif ace["Ace"]["Flags"] == 1:
                    uuid = format_uuid_le(ace["Ace"]["ObjectType"])[1:-1].lower()
                    self._logger.log(self._logger.ULTRA, "UUID found in ObjectType: {}".format(uuid))
                else:
                    continue

                if uuid in adobj.PKIEnrollmentService._enrollment_uuids.values():
                    if resolve_sids:
                        resolved_sid = self._resolve_sid(sid, sid_mapping)
                    else:
                        resolved_sid = sid
                    enrollment_principals.append(resolved_sid)
                    pki.add_attributes({'allowedprincipals': enrollment_principals})
                else:
                    self._logger.log(self._logger.ULTRA, "UUID is not a known enrollment UUID: {}".format(uuid))
                    continue
            if not full_data:
                pki._attributes_dict.pop("ntsecuritydescriptor")
            objectpki.append(pki)
        return objectpki

    @LDAPRPCRequester._ldap_connection_init
    def get_netcerttmpl(self, queried_domain=str(), resolve_sids=False, full_data=False, queried_ca_name=str()):

        # This function is mostly based on the dumpADCS() one within impacket's ntlmrelayx
        # credit goes to the multiple contributors!
        base_dn = self._base_dn
        if queried_ca_name:
            self._logger.debug('Queried CA: {}'.format(queried_ca_name))
            ldap_filter = '(&(objectClass=pKIEnrollmentService)(displayname={}))'.format(queried_ca_name)
            attributes = ["certificateTemplates"]
            self._base_dn = self._server_info.other['configurationNamingContext'][0]
            try:
                queried_templates = self._ldap_search(ldap_filter, adobj.PKIEnrollmentService,
                                    attributes=attributes)[0].certificatetemplates
                self._logger.debug('Retrieved certificat templates : {}'.format(queried_templates))
            except IndexError:
                self._logger.critical('We did not manage to find this CA, please specify a valid name')
                raise ValueError('pKIEnrollmentService {} was not found'.format(queried_ca_name))
            self._base_dn = base_dn

        if full_data:
            attributes=list()
        else:
            attributes = ["msPKI-Enrollment-Flag", "name", "nTSecurityDescriptor", "pKIExtendedKeyUsage"]

        controls = security_descriptor_control(criticality=True, sdflags=LDAP_SERVER_SD_FLAGS.DACL_SECURITY_INFORMATION.value)
        config_naming_context = 'CN=Certificate Templates,CN=Public Key Services,CN=Services,{}'.format(self._server_info.other['configurationNamingContext'][0])
        self._base_dn = config_naming_context
        ldap_filter = '(objectClass=pKICertificateTemplate)'
        object_cert_template_raw = self._ldap_search(ldap_filter, adobj.PKICertificateTemplate, attributes=attributes, controls=controls)
        self._base_dn = base_dn

        sid_mapping = adobj.ADObject._well_known_sids.copy()
        object_cert_template = list()
        for cert_template in object_cert_template_raw:
            if queried_ca_name:
                if cert_template.name not in queried_templates:
                    self._logger.debug('{0} is not a template of {1}, skipping'.format(cert_template.name, queried_ca_name))
                    continue
            sd = SR_SECURITY_DESCRIPTOR()
            sd.fromString(cert_template.ntsecuritydescriptor)

            enrollment_principals = list()
            # AccessAllowedObject = 0x05
            for ace in (a for a in sd["Dacl"]["Data"] if a["AceType"] == 0x05):
                sid = format_sid(ace["Ace"]["Sid"].getData())
                if ace["Ace"]["Flags"] == 2:
                    uuid = format_uuid_le(ace["Ace"]["InheritedObjectType"])[1:-1].lower()
                    self._logger.log(self._logger.ULTRA, "UUID found in InheritedObjectType: {}".format(uuid))
                elif ace["Ace"]["Flags"] == 1:
                    uuid = format_uuid_le(ace["Ace"]["ObjectType"])[1:-1].lower()
                    self._logger.log(self._logger.ULTRA, "UUID found in ObjectType: {}".format(uuid))
                else:
                    continue

                if uuid in adobj.PKICertificateTemplate._enrollment_uuids.values():
                    if resolve_sids:
                        resolved_sid = self._resolve_sid(sid, sid_mapping)
                    else:
                        resolved_sid = sid
                    enrollment_principals.append(resolved_sid)
                    cert_template.add_attributes({'allowedprincipals': enrollment_principals})
                else:
                    self._logger.log(self._logger.ULTRA, "UUID is not a known enrollment UUID: {}".format(uuid))
                    continue
            if not full_data:
                cert_template._attributes_dict.pop("ntsecuritydescriptor")
            object_cert_template.append(cert_template)
        return object_cert_template

    @LDAPRPCRequester._ldap_connection_init
    def get_adobject(self, queried_domain=str(), queried_sid=str(),
                     queried_name=str(), queried_sam_account_name=str(),
                     ads_path=str(), attributes=list(), custom_filter=str()):
        for attr_desc, attr_value in (('objectSid', queried_sid), ('name', escape_filter_chars(queried_name)),
                                      ('samAccountName', escape_filter_chars(queried_sam_account_name))):
            if attr_value:
                object_filter = '(&({}={}){})'.format(attr_desc, attr_value, custom_filter)
                break
        else:
            object_filter = '(&(name=*){})'.format(custom_filter)

        return self._ldap_search(object_filter, adobj.ADObject, attributes=attributes)
 
    @LDAPRPCRequester._ldap_connection_init
    def get_objectowner(self, queried_domain=str(), queried_sid=str(),
                     queried_name=str(), queried_sam_account_name=str(),
                     ads_path=str(), custom_filter=str(), resolve_sids=False):
        for attr_desc, attr_value in (('objectSid', queried_sid), ('name', escape_filter_chars(queried_name)),
                                      ('samAccountName', escape_filter_chars(queried_sam_account_name))):
            if attr_value:
                object_filter = '(&({}={}){})'.format(attr_desc, attr_value, custom_filter)
                break
        else:
            object_filter = '(&(name=*){})'.format(custom_filter)

        attributes = ['ntsecuritydescriptor','distinguishedname']

        # The control is used to get access to ntSecurityDescriptor with an
        # unprivileged user, see https://stackoverflow.com/questions/40771503/selecting-the-ad-ntsecuritydescriptor-attribute-as-a-non-admin/40773088
        # /!\ May break pagination from what I've read (see Stack Overflow answer)
        controls = security_descriptor_control(criticality=True, sdflags=LDAP_SERVER_SD_FLAGS.OWNER_SECURITY_INFORMATION.value)

        objectowners_raw = self._ldap_search(object_filter, adobj.ADObject, attributes=attributes, controls=controls)
        objectowners = list()
        if resolve_sids:
            sid_mapping = adobj.ADObject._well_known_sids.copy()

        for objectowner in objectowners_raw:
            # Sometimes, admins mess with objects, we skip if the object does not have ntsecuritydescriptor
            if not objectowner.ntsecuritydescriptor:
                self._logger.debug("Skipping an object because it does not have a ntsecuritydescriptor attributes")
                continue
            result = adobj.ObjectOwner(list())
            sd = SR_SECURITY_DESCRIPTOR()
            sd.fromString(objectowner.ntsecuritydescriptor)
            sid = format_sid(sd['OwnerSid'].getData())
            if resolve_sids:
                resolved_sid = self._resolve_sid(sid, sid_mapping)
            else:
                resolved_sid = sid
            result.add_attributes({'distinguishedname': objectowner.distinguishedname})
            result.add_attributes({'objectowner': resolved_sid})
            objectowners.append(result)
        return objectowners

    @LDAPRPCRequester._ldap_connection_init
    def get_netgmsa(self, queried_domain=str(), queried_sid=str(),
                     queried_name=str(), queried_sam_account_name=str(),
                     ads_path=str(), resolve_sids=False):
        filter_objectclass = '(ObjectClass=msDS-GroupManagedServiceAccount)'
        attributes = ['samaccountname', 'distinguishedname', 'objectsid', 'description',
                      'msds-managedpassword', 'msds-groupmsamembership', 'useraccountcontrol']

        if not self._ldap_connection.server.ssl:
            self._logger.warning('LDAP connection is not encrypted, we can\'t ask '\
                    'for msds-managedpassword, removing from list of attributes')
            attributes.remove('msds-managedpassword')

        for attr_desc, attr_value in (('objectSid', queried_sid), ('name', escape_filter_chars(queried_name)),
                                      ('samAccountName', escape_filter_chars(queried_sam_account_name))):
            if attr_value:
                object_filter = '(&({}={}){})'.format(attr_desc, attr_value, filter_objectclass)
                break
        else:
            object_filter = '(&(name=*){})'.format(filter_objectclass)

        sid_mapping = adobj.ADObject._well_known_sids.copy()
        gmsa = self._ldap_search(object_filter, adobj.GMSAAccount, attributes=attributes)

        # In this loop, we resolve SID (if true) and we populate 'enabled' attribute
        for i, adserviceaccount in enumerate(gmsa):
            if resolve_sids:
                results = list()
                for sid in getattr(adserviceaccount, 'msds-groupmsamembership'):
                    resolved_sid = self._resolve_sid(sid, sid_mapping)
                    results.append(resolved_sid)
                gmsa[i].add_attributes({'msds-groupmsamembership': results})
            gmsa[i].add_attributes({'Enabled': 'ACCOUNTDISABLE' not in adserviceaccount.useraccountcontrol})
            gmsa[i]._attributes_dict.pop('useraccountcontrol')

        return gmsa

    @LDAPRPCRequester._ldap_connection_init
    def get_netsmsa(self, queried_domain=str(), queried_sid=str(),
                     queried_name=str(), queried_sam_account_name=str(),
                     ads_path=str()):

        filter_objectclass = '(ObjectClass=msDS-ManagedServiceAccount)'
        attributes = ['samaccountname', 'distinguishedname', 'objectsid', 'description',
                      'msds-hostserviceaccountbl', 'useraccountcontrol']
        for attr_desc, attr_value in (('objectSid', queried_sid), ('name', escape_filter_chars(queried_name)),
                                      ('samAccountName', escape_filter_chars(queried_sam_account_name))):
            if attr_value:
                object_filter = '(&({}={}){})'.format(attr_desc, attr_value, filter_objectclass)
                break
        else:
            object_filter = '(&(name=*){})'.format(filter_objectclass)

        smsas = self._ldap_search(object_filter, adobj.SMSAAccount, attributes=attributes)

        # In this loop, we populate 'enabled' attribute
        for i, adserviceaccount in enumerate(smsas):
            smsas[i].add_attributes({'Enabled': 'ACCOUNTDISABLE' not in adserviceaccount.useraccountcontrol})
            smsas[i]._attributes_dict.pop('useraccountcontrol')

        return smsas

    @LDAPRPCRequester._ldap_connection_init
    def get_objectacl(self, queried_domain=str(), queried_sid=str(),
                     queried_name=str(), queried_sam_account_name=str(),
                     ads_path=str(), sacl=False, rights_filter=str(),
                     resolve_sids=False, resolve_guids=False, custom_filter=str()):
        for attr_desc, attr_value in (('objectSid', queried_sid), ('name', escape_filter_chars(queried_name)),
                                      ('samAccountName', escape_filter_chars(queried_sam_account_name))):
            if attr_value:
                object_filter = '(&({}={}){})'.format(attr_desc, attr_value, custom_filter)
                break
        else:
            object_filter = '(&(name=*){})'.format(custom_filter)

        guid_map = dict()
        # This works on a mono-domain forest, must be tested on a more complex one
        if resolve_guids:
            # Dirty fix to get base DN even if custom ADS path was given
            len_base_dn = len(self._base_dn.split(','))
            base_dn = ','.join(self._base_dn.split(',')[-len_base_dn:])
            guid_map = {'{00000000-0000-0000-0000-000000000000}': 'All'}
            for o in self.get_adobject(ads_path=self._server_info.other['schemaNamingContext'][0],
                    attributes=['name', 'schemaIDGUID'], custom_filter='(schemaIDGUID=*)'):
                        guid_map['{}'.format(format_uuid_le(o.schemaidguid))] = o.name

            for o in self.get_adobject(ads_path='CN=Extended-Rights,{}'.format(self._server_info.other['configurationNamingContext'][0]),
                    attributes=['name', 'rightsGuid'], custom_filter='(objectClass=controlAccessRight)'):
                        guid_map['{{{}}}'.format(o.rightsguid.lower())] = o.name
            self._base_dn = base_dn

        attributes = ['distinguishedname', 'objectsid', 'ntsecuritydescriptor']

        if sacl:
            controls = list()
            acl_type = 'Sacl'
        else:
            # The control is used to get access to ntSecurityDescriptor with an
            # unprivileged user, see https://stackoverflow.com/questions/40771503/selecting-the-ad-ntsecuritydescriptor-attribute-as-a-non-admin/40773088
            # /!\ May break pagination from what I've read (see Stack Overflow answer)
            sdflags = LDAP_SERVER_SD_FLAGS.OWNER_SECURITY_INFORMATION.value | \
                      LDAP_SERVER_SD_FLAGS.GROUP_SECURITY_INFORMATION.value | \
                      LDAP_SERVER_SD_FLAGS.DACL_SECURITY_INFORMATION.value
            controls = security_descriptor_control(criticality=True, sdflags=sdflags)
            acl_type = 'Dacl'

        security_descriptors = self._ldap_search(object_filter, adobj.ADObject,
                attributes=attributes, controls=controls)

        acl = list()

        rights_to_guid = {'reset-password': '{00299570-246d-11d0-a768-00aa006e0529}',
                'write-members': '{bf9679c0-0de6-11d0-a285-00aa003049e2}',
                'allowed-to-authenticate':'{68b1d179-0d15-4d4f-ab71-46152e79a7bc}',
                'all': '{00000000-0000-0000-0000-000000000000}'}
        guid_filter = rights_to_guid.get(rights_filter, None)

        if resolve_sids:
            sid_mapping = adobj.ADObject._well_known_sids.copy()

        for security_descriptor in security_descriptors:
            sd = SR_SECURITY_DESCRIPTOR()
            try:
                sd.fromString(security_descriptor.ntsecuritydescriptor)
            except TypeError:
                continue
            for ace in sd[acl_type]['Data']:
                if guid_filter:
                    try:
                        object_type = format_uuid_le(ace['Ace']['ObjectType']) if ace['Ace']['ObjectType'] else '{00000000-0000-0000-0000-000000000000}'
                    except KeyError:
                        continue
                    if object_type != guid_filter:
                        continue
                attributes = dict()
                attributes['objectdn'] = security_descriptor.distinguishedname
                attributes['objectsid'] = security_descriptor.objectsid
                attributes['acetype'] = ace['TypeName']
                attributes['binarysize'] = ace['AceSize']
                attributes['aceflags'] = fmt.format_ace_flags(ace['AceFlags'])
                attributes['accessmask'] = ace['Ace']['Mask']['Mask']
                attributes['activedirectoryrights'] = fmt.format_ace_access_mask(ace['Ace']['Mask']['Mask'])
                attributes['isinherited'] = bool(ace['AceFlags'] & 0x10)
                attributes['securityidentifier'] = format_sid(ace['Ace']['Sid'].getData())
                if resolve_sids:
                    converted_sid = attributes['securityidentifier']
                    attributes['securityidentifier'] = self._resolve_sid(converted_sid, sid_mapping)
                try:
                    attributes['objectaceflags'] = fmt.format_object_ace_flags(ace['Ace']['Flags'])
                except KeyError:
                    pass
                try:
                    attributes['objectacetype'] = format_uuid_le(ace['Ace']['ObjectType']) if ace['Ace']['ObjectType'] else '{00000000-0000-0000-0000-000000000000}'
                    attributes['objectacetype'] = guid_map[attributes['objectacetype']]
                except KeyError:
                    pass
                try:
                    attributes['inheritedobjectacetype'] = format_uuid_le(ace['Ace']['InheritedObjectType']) if ace['Ace']['InheritedObjectType'] else '{00000000-0000-0000-0000-000000000000}'
                    attributes['inheritedobjectacetype'] = guid_map[attributes['inheritedobjectacetype']]
                except KeyError:
                    pass

                acl.append(adobj.ACE(attributes))

        return acl

    @LDAPRPCRequester._ldap_connection_init
    def get_netuser(self, queried_username=str(), queried_domain=str(),
                    ads_path=str(), admin_count=False, spn=False,
                    unconstrained=False, allow_delegation=False,
                    preauth_notreq=False,
                    custom_filter=str(), attributes=[]):

        if unconstrained:
            custom_filter += '(userAccountControl:1.2.840.113556.1.4.803:=524288)'

        if allow_delegation:
            custom_filter += '(!(userAccountControl:1.2.840.113556.1.4.803:=1048574))'

        if admin_count:
            custom_filter += '(admincount=1)'
        # LDAP filter from https://www.harmj0y.net/blog/activedirectory/roasting-as-reps/
        if preauth_notreq:
            custom_filter += '(userAccountControl:1.2.840.113556.1.4.803:=4194304)'
        user_search_filter = '(samAccountType=805306368){}'.format(custom_filter)
        if queried_username:
            user_search_filter += '(samAccountName={})'.format(queried_username)
        elif spn:
            user_search_filter += '(servicePrincipalName=*)'

        user_search_filter = '(&{})'.format(user_search_filter)

        return self._ldap_search(user_search_filter, adobj.User, attributes=attributes)

    @LDAPRPCRequester._ldap_connection_init
    def get_netgroup(self, queried_groupname='*', queried_sid=str(),
                     queried_username=str(), queried_domain=str(),
                     ads_path=str(), admin_count=False, full_data=False,
                     custom_filter=str()):

        # RFC 4515, section 3
        # However if we escape *, we can no longer use wildcard within `--groupname`
        if not '*' in queried_groupname:
            queried_groupname = escape_filter_chars(queried_groupname)
        else:
            self._logger.warning('"*" detected in "{}", if it also contains "(",")" or "\\", '
                                 'script will probably crash ("invalid filter"). '
                                 'Don\'t use wildcard with these characters'.format(queried_groupname))

        if queried_username:
            self._logger.debug('Queried username = {}'.format(queried_username))
            results = list()
            sam_account_name_to_resolve = [queried_username]
            first_run = True
            while sam_account_name_to_resolve:
                sam_account_name = escape_filter_chars(sam_account_name_to_resolve.pop(0))
                if first_run:
                    first_run = False
                    if admin_count:
                        custom_filter = '(&{}(admincount=1))'.format(custom_filter)
                    objects = self.get_adobject(queried_sam_account_name=sam_account_name,
                                                queried_domain=queried_domain,
                                                ads_path=ads_path, custom_filter=custom_filter)
                    objects += self.get_adobject(queried_name=sam_account_name,
                                                 queried_domain=queried_domain,
                                                 ads_path=ads_path, custom_filter=custom_filter)
                else:
                    objects = self.get_adobject(queried_sam_account_name=sam_account_name,
                                                queried_domain=queried_domain)
                    objects += self.get_adobject(queried_name=sam_account_name,
                                                 queried_domain=queried_domain)

                for obj in objects:
                    try:
                        if not isinstance(obj.memberof, list):
                            obj.memberof = [obj.memberof]
                    except AttributeError:
                        continue
                    for group_dn in obj.memberof:
                        group_sam_account_name = group_dn.split(',')[0].split('=')[1]
                        if not group_sam_account_name in results:
                            results.append(group_sam_account_name)
                            sam_account_name_to_resolve.append(group_sam_account_name)
            final_results = list()
            for group_sam_account_name in results:
                obj_member_of = adobj.Group(list())
                obj_member_of._attributes_dict['samaccountname'] = group_sam_account_name
                final_results.append(obj_member_of)
            return final_results
        else:
            if admin_count:
                custom_filter += '(admincount=1)'

            group_search_filter = custom_filter
            group_search_filter += '(objectCategory=group)'

            if queried_sid:
                self._logger.debug('Queried SID = {}'.format(queried_username))
                group_search_filter += '(objectSid={})'.format(queried_sid)
            elif queried_groupname:
                self._logger.debug('Queried groupname = {}'.format(queried_groupname))
                group_search_filter += '(name={})'.format(queried_groupname)

            if full_data:
                attributes=list()
            else:
                attributes=['samaccountname']

            group_search_filter = '(&{})'.format(group_search_filter)
            return self._ldap_search(group_search_filter, adobj.Group, attributes=attributes)

    @LDAPRPCRequester._ldap_connection_init
    def get_netcomputer(self, queried_computername=str(), queried_spn=str(),
                        queried_os=str(), queried_sp=str(), queried_domain=str(), ads_path=str(), 
                        printers=False, unconstrained=False, laps_passwords=False, pre_created=False,
                        ping=False, full_data=False, custom_filter=str(), attributes=[]):

        if unconstrained:
            custom_filter += '(userAccountControl:1.2.840.113556.1.4.803:=524288)'

        if printers:
            custom_filter += '(objectCategory=printQueue)'

        if laps_passwords:
            custom_filter += '(ms-mcs-AdmPwd=*)'

        if pre_created:
            custom_filter += '(userAccountControl:1.2.840.113556.1.4.803:=4128)'

        computer_search_filter = '(samAccountType=805306369){}'.format(custom_filter)
        for (attr_desc, attr_value) in (('servicePrincipalName', queried_spn),
                ('operatingSystem', queried_os), ('operatingsystemservicepack', queried_sp),
                ('dnsHostName', queried_computername)):
            if attr_value:
                computer_search_filter += '({}={})'.format(attr_desc, attr_value)

        if full_data:
            attributes=list()
        else:
            if not attributes:
                attributes=['samaccountname', 'dnsHostName']
            if laps_passwords:
                attributes.append('ms-mcs-AdmPwd')

        computer_search_filter = '(&{})'.format(computer_search_filter)

        return self._ldap_search(computer_search_filter, adobj.Computer, attributes=attributes)

    @LDAPRPCRequester._ldap_connection_init
    def get_netdomaincontroller(self, queried_domain=str()):

        domain_controller_filter = '(userAccountControl:1.2.840.113556.1.4.803:=8192)'

        return self.get_netcomputer(queried_domain=queried_domain, full_data=True,
                                    custom_filter=domain_controller_filter)

    @LDAPRPCRequester._ldap_connection_init
    def get_netfileserver(self, queried_domain=str(), target_users=list()):

        def split_path(path):
            split_path = path.split('\\')
            if len(split_path) >= 3:
                return split_path[2]

        file_server_attributes = ['homedirectory', 'scriptpath', 'profilepath']
        results = set()
        if target_users:
            users = list()
            for target_user in target_users:
                users += self.get_netuser(target_user, queried_domain,
                        attributes=file_server_attributes)
        else:
            users = self.get_netuser(queried_domain=queried_domain,
                    attributes=file_server_attributes)

        for user in users:
            for full_path in (user.homedirectory, user.scriptpath, user.profilepath):
                if not full_path:
                    continue
                path = split_path(full_path)
                if path:
                    results.add(path)

        final_results = list()
        for file_server_name in results:
            attributes = dict()
            attributes['dnshostname'] = file_server_name
            final_results.append(adobj.FileServer(attributes))

        return final_results

    @LDAPRPCRequester._ldap_connection_init
    def get_dfsshare(self, version=['v1', 'v2'], queried_domain=str(), ads_path=str()):

        def _get_dfssharev1():
            dfs_search_filter = '(objectClass=fTDfs)'

            intermediate_results = self._ldap_search(dfs_search_filter, adobj.ADObject,
                                                attributes=['remoteservername', 'name'])
            results = list()
            for dfs in intermediate_results:
                for remote_server in dfs.remoteservername:
                    remote_server = str(remote_server)
                    if '\\' in remote_server:
                        attributes = {'name': dfs.name,
                                'remoteservername': remote_server.split('\\')[2]}
                        results.append(adobj.DFS(attributes))

            return results

        def _get_dfssharev2():
            dfs_search_filter = '(objectClass=msDFS-Linkv2)'

            intermediate_results = self._ldap_search(dfs_search_filter, adobj.ADObject,
                                                attributes=['msdfs-linkpathv2','msDFS-TargetListv2'])
            results = list()
            for dfs in intermediate_results:
                attributes = list()

                share_name = getattr(dfs, 'msdfs-linkpathv2')

                xml_target_list = getattr(dfs, 'msdfs-targetlistv2')[2:].decode('utf-16le')
                soup_target_list = BeautifulSoup(xml_target_list, 'xml')
                for target in soup_target_list.targets.contents:
                    if '\\' in target.string:
                        server_name, dfs_root = target.string.split('\\')[2:4]
                        attributes = {'name': '{}{}'.format(dfs_root, share_name),
                                'remoteservername': server_name}

                results.append(adobj.DFS(attributes))

            return results

        version_to_function = {'v1': _get_dfssharev1, 'v2': _get_dfssharev2}
        results = list()

        for v in version:
            results += version_to_function[v]()

        return results

    @LDAPRPCRequester._ldap_connection_init
    def get_netou(self, queried_domain=str(), queried_ouname='*',
                  queried_guid=str(), ads_path=str(), full_data=False):

        ou_search_filter = '(objectCategory=organizationalUnit)'

        if queried_ouname:
            ou_search_filter += '(name={})'.format(queried_ouname)

        if queried_guid:
            ou_search_filter += '(gplink=*{}*)'.format(queried_guid)

        if full_data:
            attributes = list()
        else:
            attributes = ['distinguishedName']

        ou_search_filter = '(&{})'.format(ou_search_filter)

        return self._ldap_search(ou_search_filter, adobj.OU, attributes=attributes)

    @LDAPRPCRequester._ldap_connection_init
    def get_netsite(self, queried_domain=str(), queried_sitename=str(),
                    queried_guid=str(), ads_path=str(), ads_prefix=str(),
                    full_data=False):

        site_search_filter = '(objectCategory=site)'

        if queried_sitename:
            site_search_filter += '(name={})'.format(queried_sitename)

        if queried_guid:
            site_search_filter += '(gplink=*{}*)'.format(queried_guid)

        if full_data:
            attributes = list()
        else:
            attributes = ['name']

        site_search_filter = '(&{})'.format(site_search_filter)

        return self._ldap_search(site_search_filter, adobj.Site, attributes=attributes)

    @LDAPRPCRequester._ldap_connection_init
    def get_netsubnet(self, queried_domain=str(), queried_sitename=str(),
                      ads_path=str(), ads_prefix=str(), full_data=False):

        subnet_search_filter = '(objectCategory=subnet)'

        if queried_sitename:
            if not queried_sitename.endswith('*'):
                queried_sitename += '*'
            subnet_search_filter += '(siteobject=*CN={})'.format(queried_sitename)

        if full_data:
            attributes = list()
        else:
            attributes = ['name', 'siteobject']

        subnet_search_filter = '(&{})'.format(subnet_search_filter)

        return self._ldap_search(subnet_search_filter, adobj.Subnet, attributes=attributes)

    @LDAPRPCRequester._ldap_connection_init
    def get_netgroupmember(self, queried_groupname=str(), queried_sid=str(),
                           queried_domain=str(), ads_path=str(), recurse=False,
                           use_matching_rule=False, full_data=False,
                           custom_filter=str()):

        def _get_members(_groupname=str(), _sid=str()):
            try:
                # `--groupname` option is supplied
                if _groupname:
                    self._logger.debug('Queried groupname = {}'.format(queried_groupname))
                    groups = self.get_netgroup(queried_groupname=_groupname,
                                               queried_domain=self._queried_domain,
                                               full_data=True)

                # `--groupname` option is missing, falling back to the "Domain Admins"
                else:
                    self._logger.debug('No groupname provided, falling back to the "Domain Admins"'.format(queried_groupname))
                    if _sid:
                        queried_sid = _sid
                    else:
                        # Logic extract from pywerview.functions.Misc get_domainsid to save object creation
                        # LDAP filter to extract DC
                        domain_controller_filter = '(userAccountControl:1.2.840.113556.1.4.803:=8192)'
                        domain_controllers = self.get_netcomputer(queried_domain=queried_domain, custom_filter=domain_controller_filter,
                                                                attributes=['objectsid'])
                        if domain_controllers:
                            primary_dc = domain_controllers[0]
                            domain_sid = primary_dc.objectsid

                            # we need to retrieve the domain sid from the controller sid
                            domain_sid = '-'.join(domain_sid.split('-')[:-1])
                            queried_sid = domain_sid + '-512'
                            self._logger.debug('Found Domains Admins SID = {}'.format(queried_sid))
                        else:
                            self._logger.critical('We did not manage to retrieve domain controller, please specify a group name')
                            return list()

                    groups = self.get_netgroup(queried_sid=queried_sid,
                                               queried_domain=self._queried_domain,
                                               full_data=True)
            except IndexError:
                raise ValueError('The group {} was not found'.format(_groupname))

            final_members = list()

            for group in groups:
                members = list()
                if recurse and use_matching_rule:
                    group_memberof_filter = '(&(samAccountType=805306368)(memberof:1.2.840.113556.1.4.1941:={}){})'.format(group.distinguishedname, custom_filter)

                    members = self.get_netuser(custom_filter=group_memberof_filter,
                                               queried_domain=self._queried_domain)
                else:
                    # TODO: range cycling
                    try:
                        for member in group.member:
                            # RFC 4515, section 3
                            self._logger.warning('Member name = "{}" will be escaped'.format(member))
                            member = escape_filter_chars(member, encoding='utf-8')
                            dn_filter = '(distinguishedname={}){}'.format(member, custom_filter)
                            result = self.get_adobject(custom_filter=dn_filter, queried_domain=self._queried_domain)
                            if not result:
                                self._logger.warning('{} not found on this domain, it may be from another domain or no longer exist'.format(member))
                                # We create a dummy object
                                result = [adobj.ADObject({'distinguishedname':member})]
                            members += result
                    # The group doesn't have any members
                    except AttributeError:
                        self._logger.debug('The group doesn\'t have any members')
                        continue

                for member in members:
                    if full_data:
                        final_member = member
                    else:
                        final_member = adobj.ADObject(list())

                    member_dn = member.distinguishedname
                    try:
                        member_domain = member_dn[member_dn.index('DC='):].replace('DC=', '').replace(',', '.')
                    except IndexError:
                        self._logger.warning('Exception was raised while handling member_dn, falling back to empty string')
                        member_domain = str()

                    attributes = dict()

                    # Member is from another domain or no longer exist
                    try:
                        member.objectsid
                    except AttributeError:
                        self._logger.warning('Member is maybe from another domain')
                        if queried_domain:
                            attributes['groupdomain'] = queried_domain
                        else:
                            attributes['groupdomain'] = self._queried_domain
                        attributes['groupname'] = group.name
                        attributes['membername'] = str()
                        attributes['memberdomain'] = str()
                        attributes['useraccountcontrol'] = str()
                        attributes['isgroup'] = bool()
                        attributes['memberdn'] = member_dn
                        attributes['objectsid'] = str()
                        final_member.add_attributes(attributes)
                        final_members.append(final_member)
                        continue

                    # Member is a Foreign Security Principal
                    if 'foreignSecurityPrincipal' in member.objectclass:
                        self._logger.warning('Member is a Foreign Security Principal')
                        if queried_domain:
                            attributes['groupdomain'] = queried_domain
                        else:
                            attributes['groupdomain'] = self._queried_domain
                        attributes['groupname'] = group.name
                        attributes['membername'] = str()
                        attributes['memberdomain'] = str()
                        attributes['useraccountcontrol'] = str()
                        attributes['isgroup'] = bool()
                        attributes['memberdn'] = member_dn
                        attributes['objectsid'] = member.objectsid
                        final_member.add_attributes(attributes)
                        final_members.append(final_member)
                        continue

                    # https://serverfault.com/questions/788888/what-does-an-alias-group-means-in-sid-context
                    is_group = (member.samaccounttype == 'GROUP_OBJECT') or (member.samaccounttype == 'ALIAS_OBJECT')

                    if queried_domain:
                        attributes['groupdomain'] = queried_domain
                    else:
                        attributes['groupdomain'] = self._queried_domain
                    attributes['groupname'] = group.name
                    attributes['membername'] = member.samaccountname
                    attributes['memberdomain'] = member_domain
                    if is_group:
                        attributes['useraccountcontrol'] = str()
                        self._logger.debug('{0} is a {1}, ignoring the useraccountcontrol'.format(member.samaccountname, member.samaccounttype))
                    else:
                        attributes['useraccountcontrol'] = member.useraccountcontrol
                    attributes['isgroup'] = is_group
                    attributes['memberdn'] = member_dn
                    attributes['objectsid'] = member.objectsid

                    final_member.add_attributes(attributes)

                    final_members.append(final_member)

            return final_members

        results = list()
        groups_to_process = [(queried_groupname, queried_sid)]

        while groups_to_process:
            groupname, sid = groups_to_process.pop(0)
            members = _get_members(groupname, sid)

            for member in members:
                results.append(member)
                if (recurse and (not use_matching_rule) and member.isgroup and member.membername):
                    groups_to_process.append((member.membername, str()))

        return results

    @LDAPRPCRequester._ldap_connection_init
    def get_netdomaintrust(self, queried_domain, full_data=False):

        if full_data:
            attributes=list()
        else:
            attributes=['trustpartner', 'trustdirection', 'whencreated', 'whenchanged',
                        'trusttype', 'trustattributes', 'securityidentifier']

        trust_search_filter = '(&(objectClass=trustedDomain))'

        return self._ldap_search(trust_search_filter, adobj.Trust, attributes=attributes)

    @LDAPRPCRequester._rpc_connection_init(r'\srvsvc')
    def get_netsession(self):

        try:
            resp = srvs.hNetrSessionEnum(self._rpc_connection, '\x00', NULL, 10)
        except DCERPCException:
            return list()

        results = list()
        for session in resp['InfoStruct']['SessionInfo']['Level10']['Buffer']:
            results.append(rpcobj.Session(session))

        return results

    @LDAPRPCRequester._rpc_connection_init(r'\srvsvc')
    def get_netshare(self):

        resp = srvs.hNetrShareEnum(self._rpc_connection, 1)

        results = list()
        for share in resp['InfoStruct']['ShareInfo']['Level1']['Buffer']:
            results.append(rpcobj.Share(share))

        return results

    @LDAPRPCRequester._rpc_connection_init(r'\srvsvc')
    def get_localdisks(self):

        resp = srvs.hNetrServerDiskEnum(self._rpc_connection, 0)

        results = list()
        for disk in resp['DiskInfoStruct']['Buffer']:
            if disk['Disk'] != '\x00':
                results.append(rpcobj.Disk(disk))

        return results

    @LDAPRPCRequester._rpc_connection_init(r'\samr')
    def get_netdomain(self):

        resp = samr.hSamrConnect(self._rpc_connection)
        server_handle = resp['ServerHandle']

        # We first list every domain in the SAM
        resp = samr.hSamrEnumerateDomainsInSamServer(self._rpc_connection, server_handle)

        results = list()
        for domain in resp['Buffer']['Buffer']:
            results.append(domain['Name'])

        return results

    @LDAPRPCRequester._rpc_connection_init(r'\wkssvc')
    def get_netloggedon(self):

        try:
            resp = wkst.hNetrWkstaUserEnum(self._rpc_connection, 1)
        except DCERPCException:
            return list()

        results = list()
        for wksta_user in resp['UserInfo']['WkstaUserInfo']['Level1']['Buffer']:
            results.append(rpcobj.WkstaUser(wksta_user))

        return results

    # TODO: if self._target_computer == self._domain_controller, check that
    # self._domain_controller is indeed a domain controller
    @LDAPRPCRequester._ldap_connection_init
    @LDAPRPCRequester._rpc_connection_init(r'\samr')
    def get_netlocalgroup(self, queried_groupname=str(), list_groups=False,
                          recurse=False):
        from impacket.nt_errors import STATUS_MORE_ENTRIES
        results = list()

        resp = samr.hSamrConnect(self._rpc_connection)
        server_handle = resp['ServerHandle']

        # We first list every domain in the SAM
        resp = samr.hSamrEnumerateDomainsInSamServer(self._rpc_connection, server_handle)
        domains = resp['Buffer']['Buffer']
        domain_handles = dict()
        for local_domain in domains:
            resp = samr.hSamrLookupDomainInSamServer(self._rpc_connection, server_handle, local_domain['Name'])
            domain_sid = 'S-1-5-{}'.format('-'.join(str(x) for x in resp['DomainId']['SubAuthority']))
            resp = samr.hSamrOpenDomain(self._rpc_connection, serverHandle=server_handle, domainId=resp['DomainId'])
            domain_handles[domain_sid] = resp['DomainHandle']

        # If we list the groups
        if list_groups:
            # We browse every domain
            for domain_sid, domain_handle in domain_handles.items():
                # We enumerate local groups in every domain
                enumeration_context = 0
                groups = list()
                while True:
                    resp = samr.hSamrEnumerateAliasesInDomain(self._rpc_connection, domain_handle,
                            enumerationContext=enumeration_context)
                    groups += resp['Buffer']['Buffer']

                    enumeration_context = resp['EnumerationContext']
                    if resp['ErrorCode'] != STATUS_MORE_ENTRIES:
                        break

                # We get information on every group
                for group in groups:
                    resp = samr.hSamrRidToSid(self._rpc_connection, domain_handle, rid=group['RelativeId'])
                    sid = 'S-1-5-{}'.format('-'.join(str(x) for x in resp['Sid']['SubAuthority']))

                    resp = samr.hSamrOpenAlias(self._rpc_connection, domain_handle, aliasId=group['RelativeId'])
                    alias_handle = resp['AliasHandle']
                    resp = samr.hSamrQueryInformationAlias(self._rpc_connection, alias_handle)

                    final_group = rpcobj.Group(resp['Buffer']['General'])
                    final_group.add_attributes({'server': self._target_computer, 'sid': sid})

                    results.append(final_group)

                    samr.hSamrCloseHandle(self._rpc_connection, alias_handle)

                samr.hSamrCloseHandle(self._rpc_connection, domain_handle)
        # If we query a group
        else:
            queried_group_rid = None
            queried_group_domain_handle = None

            # If the user is looking for a particular group
            if queried_groupname:
                # We look for it in every domain
                for _, domain_handle in domain_handles.items():
                    try:
                        resp = samr.hSamrLookupNamesInDomain(self._rpc_connection, domain_handle, [queried_groupname])
                        queried_group_rid = resp['RelativeIds']['Element'][0]['Data']
                        queried_group_domain_handle = domain_handle
                        break
                    except (DCERPCSessionError, KeyError, IndexError):
                        continue
                else:
                    raise ValueError('The group \'{}\' was not found on the target server'.format(queried_groupname))
            # Otherwise, we look for the local Administrators group
            else:
                queried_group_rid = 544
                resp = samr.hSamrLookupDomainInSamServer(self._rpc_connection, server_handle, 'BUILTIN')
                resp = samr.hSamrOpenDomain(self._rpc_connection, serverHandle=server_handle, domainId=resp['DomainId'])
                queried_group_domain_handle = resp['DomainHandle']

            # We get a handle on the group, and list its members
            try:
                group = samr.hSamrOpenAlias(self._rpc_connection, queried_group_domain_handle, aliasId=queried_group_rid)
                resp = samr.hSamrGetMembersInAlias(self._rpc_connection, group['AliasHandle'])
            except DCERPCSessionError:
                raise ValueError('The name \'{}\' is not a valid group on the target server'.format(queried_groupname))

            # For every user, we look for information in every local domain
            for member in resp['Members']['Sids']:
                attributes = dict()
                member_rid = member['SidPointer']['SubAuthority'][-1]
                member_sid = 'S-1-5-{}'.format('-'.join(str(x) for x in member['SidPointer']['SubAuthority']))

                attributes['server'] = self._target_computer
                attributes['sid'] = member_sid

                for domain_sid, domain_handle in domain_handles.items():
                    # We've found a local member
                    if member_sid.startswith(domain_sid):
                        attributes['isdomain'] = False
                        resp = samr.hSamrQueryInformationDomain(self._rpc_connection, domain_handle)
                        member_domain = resp['Buffer']['General2']['I1']['DomainName']
                        try:
                            resp = samr.hSamrOpenUser(self._rpc_connection, domain_handle, userId=member_rid)
                            member_handle = resp['UserHandle']
                            attributes['isgroup'] = False
                            resp = samr.hSamrQueryInformationUser(self._rpc_connection, member_handle)
                            attributes['name'] = '{}\\{}'.format(member_domain, resp['Buffer']['General']['UserName'])
                        except DCERPCSessionError:
                            resp = samr.hSamrOpenAlias(self._rpc_connection, domain_handle, aliasId=member_rid)
                            member_handle = resp['AliasHandle']
                            attributes['isgroup'] = True
                            resp = samr.hSamrQueryInformationAlias(self._rpc_connection, member_handle)
                            attributes['name'] = '{}\\{}'.format(member_domain, resp['Buffer']['General']['Name'])
                        attributes['lastlogon'] = str()
                        break
                # It's a domain member
                else:
                    attributes['isdomain'] = True
                    if self._ldap_connection is not None:
                        try:
                            ad_object = self.get_adobject(queried_sid=member_sid)[0]
                            member_dn = ad_object.distinguishedname
                            member_domain = member_dn[member_dn.index('DC='):].replace('DC=', '').replace(',', '.')
                            try:
                                attributes['name'] = '{}\\{}'.format(member_domain, ad_object.samaccountname)
                            except AttributeError:
                                # Here, the member is a foreign security principal
                                # TODO: resolve it properly
                                self._logger.warning('The member is a foreign security principal, SID will not be resolved')
                                attributes['name'] = '{}\\{}'.format(member_domain, ad_object.objectsid)
                            attributes['isgroup'] = 'group' in ad_object.objectclass
                            try:
                                attributes['lastlogon'] = ad_object.lastlogon
                            except AttributeError:
                                self._logger.warning('lastlogon is not set, falling back to empty string')
                                attributes['lastlogon'] = str()
                        except IndexError:
                            # We did not manage to resolve this SID against the DC
                            self._logger.warning('We did not manage to resolve this SID ({}) against the DC'.format(member_sid))
                            attributes['isdomain'] = False
                            attributes['isgroup'] = False
                            attributes['name'] = attributes['sid']
                            attributes['lastlogon'] = str()
                    else:
                        attributes['isgroup'] = False
                        attributes['name'] = str()
                        attributes['lastlogon'] = str()

                results.append(rpcobj.RPCObject(attributes))

                # If we recurse and the member is a domain group, we query every member
                # TODO: implement check on self._domain_controller here?
                if self._ldap_connection and self._domain_controller and recurse and attributes['isdomain'] and attributes['isgroup']:
                    for domain_member in self.get_netgroupmember(full_data=True, recurse=True, queried_sid=attributes['sid']):
                        domain_member_attributes = dict()
                        domain_member_attributes['isdomain'] = True
                        member_dn = domain_member.distinguishedname
                        member_domain = member_dn[member_dn.index('DC='):].replace('DC=', '').replace(',', '.')
                        domain_member_attributes['name'] = '{}\\{}'.format(member_domain, domain_member.samaccountname)
                        domain_member_attributes['isgroup'] = domain_member.isgroup
                        domain_member_attributes['isdomain'] = True
                        # TODO: Nope, maybe here we can call get-netdomaincontroller ?
                        # Need to check in powerview
                        domain_member_attributes['server'] = attributes['name']
                        domain_member_attributes['sid'] = domain_member.objectsid
                        try:
                            domain_member_attributes['lastlogin'] = ad_object.lastlogon
                        except AttributeError:
                            self._logger.warning('lastlogon is not set, falling back to empty string')
                            domain_member_attributes['lastlogin'] = str()
                        results.append(rpcobj.RPCObject(domain_member_attributes))

        return results

    @LDAPRPCRequester._wmi_connection_init()
    def get_netprocess(self):
        wmi_enum_process = self._wmi_connection.ExecQuery('SELECT * from Win32_Process',
                                                          lFlags=WBEM_FLAG_FORWARD_ONLY)
        while True:
            try:
                # TODO: do we have to get them one by one?
                wmi_process = wmi_enum_process.Next(0xffffffff, 1)[0]
                wmi_process_owner = wmi_process.GetOwner()

                # Sometimes GetOwner() returns None but the list is not over
                if wmi_process_owner != None:
                    attributes = {'computername': self._target_computer,
                                  'processname': wmi_process.Name,
                                  'processid': wmi_process.ProcessId,
                                  'user': wmi_process_owner.User,
                                  'domain': wmi_process_owner.Domain}

                    result_process = rpcobj.Process(attributes)
                    yield result_process
            except Exception as e:
                if str(e).find('S_FALSE') < 0:
                    raise e
                else:
                    break

    @LDAPRPCRequester._wmi_connection_init()
    def get_userevent(self, event_type=['logon', 'tgt'], date_start=5):
        limit_date = (datetime.today() - timedelta(days=date_start)).strftime('%Y%m%d%H%M%S.%f-000')
        if event_type == ['logon']:
            where_clause = 'EventCode=4624'
        elif event_type == ['tgt']:
            where_clause = 'EventCode=4768'
        else:
            where_clause = '(EventCode=4624 OR EventCode=4768)'

        wmi_enum_event = self._wmi_connection.ExecQuery('SELECT * from Win32_NTLogEvent where {}'\
                                                        'and TimeGenerated >= \'{}\''.format(where_clause, limit_date),
                                                        lFlags=WBEM_FLAG_FORWARD_ONLY)
        while True:
            try:
                # TODO: do we have to get them one by one?
                wmi_event = wmi_enum_event.Next(0xffffffff, 1)[0]
                wmi_event_type = wmi_event.EventIdentifier
                wmi_event_info = wmi_event.InsertionStrings
                time = datetime.strptime(wmi_event.TimeGenerated, '%Y%m%d%H%M%S.%f-000')
                if wmi_event_type == 4624:
                    logon_type = int(wmi_event_info[8])
                    user = wmi_event_info[5]
                    domain = wmi_event_info[6]
                    address = wmi_event_info[18]
                    if logon_type not in [2, 3] or user.endswith('$') \
                       or (user.lower == 'anonymous logon'):
                        continue
                else:
                    logon_type = str()
                    user = wmi_event_info[0]
                    domain = wmi_event_info[1]
                    address = wmi_event_info[9].replace('::ffff:', '')

                attributes = {'computername': self._target_computer,
                              'logontype': logon_type,
                              'username': user,
                              'domain': domain,
                              'address': address,
                              'time': time,
                              'id': wmi_event_type}
                result_event = rpcobj.Event(attributes)
                yield result_event
            except Exception as e:
                if str(e).find('S_FALSE') < 0:
                    raise e
                else:
                    break




# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

import sys
import logging
import socket
import ntpath
import ldap3
import os
import tempfile
import ssl

from ldap3.protocol.formatters.formatters import *

from impacket.smbconnection import SMBConnection
from impacket.smbconnection import SessionError
from impacket.krb5.ccache import CCache, Credential, CountedOctetString
from impacket.krb5 import constants
from impacket.krb5.types import Principal
from impacket.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_PKT_PRIVACY, RPC_C_AUTHN_LEVEL_PKT_INTEGRITY
from impacket.dcerpc.v5 import transport, wkst, srvs, samr, scmr, drsuapi, epm
from impacket.dcerpc.v5.dcom import wmi
from impacket.dcerpc.v5.dtypes import NULL
from impacket.dcerpc.v5.dcomrt import DCOMConnection
from impacket.dcerpc.v5.rpcrt import DCERPCException

import pywerview.formatters as fmt

class LDAPRequester():
    def __init__(self, domain_controller, domain=str(), user=str(), password=str(),
                 lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                 do_tls=False, user_cert=str(), user_key=str()):
        self._domain_controller = domain_controller
        self._domain = domain
        self._user = user
        self._password = password
        self._lmhash = lmhash
        self._nthash = nthash
        self._do_simple = do_simple
        self._do_kerberos = do_kerberos
        self._do_certificate = user_cert and user_key
        self._user_cert = user_cert
        self._user_key = user_key
        self._do_tls = do_tls
        self._queried_domain = None
        self._ads_path = None
        self._ads_prefix = None
        self._ldap_connection = None
        self._base_dn = None
        self._server_info = None

        # Call custom formatters for several AD attributes
        self._formatter = {'userAccountControl': fmt.format_useraccountcontrol,
                'sAMAccountType': fmt.format_samaccounttype,
                'trustType': fmt.format_trusttype,
                'trustDirection': fmt.format_trustdirection,
                'trustAttributes': fmt.format_trustattributes,
                'msDS-MaximumPasswordAge': format_ad_timedelta,
                'msDS-MinimumPasswordAge': format_ad_timedelta,
                'msDS-LockoutDuration': format_ad_timedelta,
                'msDS-LockoutObservationWindow': format_ad_timedelta,
                'msDS-GroupMSAMembership': fmt.format_groupmsamembership,
                'msDS-ManagedPassword': fmt.format_managedpassword,
                'ms-Mcs-AdmPwdExpirationTime': format_ad_timestamp,
                'securityIdentifier': format_sid,
                'pKIExtendedKeyUsage': fmt.format_ekus,
                'msPKI-Enrollment-Flag': fmt.format_mspkienrollmentflag}

        logger = logging.getLogger('pywerview_main_logger.LDAPRequester')
        logger.ULTRA = 5
        self._logger = logger

        # We check here if the ldap3 install supports NTLM and Kerberos encryption
        # along with TLS channel binding
        # Run `pip install ldap3-bleeding-edge` to install it
        try:
            if ldap3.SIGN and ldap3.ENCRYPT:
                self._sign_and_seal_supported = True
                self._logger.debug('LDAP sign and seal are supported')
        except AttributeError:
            self._sign_and_seal_supported = False
            self._logger.debug('LDAP sign and seal are not supported')

        try:
            if ldap3.TLS_CHANNEL_BINDING:
                self._tls_channel_binding_supported = True
                self._logger.debug('TLS channel binding is supported')
        except AttributeError:
            self._tls_channel_binding_supported = False
            self._logger.debug('TLS channel binding is not supported')

    def _parse_ldap_invalid_credentials_result_message(self, message):
        # https://ldapwiki.com/wiki/Wiki.jsp?page=Common%20Active%20Directory%20Bind%20Errors
        error_codes = {
                        0x52e      : "ERROR_LOGON_FAILURE",
                        0x52f      : "ERROR_ACCOUNT_RESTRICTION",
                        0x530      : "ERROR_INVALID_LOGON_HOURS",
                        0x531      : "ERROR_INVALID_WORKSTATION",
                        0x532      : "ERROR_PASSWORD_EXPIRED",
                        0x533      : "ERROR_ACCOUNT_DISABLED",
                        0x701      : "ERROR_ACCOUNT_EXPIRED",
                        0x773      : "ERROR_PASSWORD_MUST_CHANGE",
                        0x775      : "ERROR_ACCOUNT_LOCKED_OUT",
                        0x80090346 : "SEC_E_BAD_BINDINGS"
                      }
        try:
            # Message is something like:
            # 80090308: LdapErr: DSID-0C090569, comment: AcceptSecurityContext error, data 775, v4563
            # we need to extract the "data" code
            data_code = message.split(": ")[3].split(', ')[1].strip(" data")
            data_code = int(data_code,16)
            parsed_code = error_codes[data_code]
        except(IndexError, KeyError):
            self._logger.warning('Unable to parse error code')
            return message
        return parsed_code

    def _do_ntlm_auth(self, server, seal_and_sign=False, tls_channel_binding=False):
        self._logger.debug('LDAP authentication with NTLM: ldap_scheme = {0} / seal_and_sign = {1} / tls_channel_binding = {2}'.format(
                            "ldaps" if server.ssl else "ldap", seal_and_sign, tls_channel_binding))
        ldap_connection_kwargs = {'server': server, 'user': '{}\\{}'.format(self._domain, self._user),
                                  'raise_exceptions': True, 'authentication': ldap3.NTLM}

        if self._lmhash or self._nthash:
            if not self._lmhash: self._lmhash = 'aad3b435b51404eeaad3b435b51404ee'
            if not self._nthash: self._nthash = '31d6cfe0d16ae931b73c59d7e0c089c0'
            ldap_connection_kwargs['password'] = '{}:{}'.format(self._lmhash, self._nthash)
        else:
            ldap_connection_kwargs['password'] = self._password
        self._logger.debug('LDAP binding parameters: server = {0} / user = {1} '
            '/ password = {2}'.format(self._domain_controller, self._user, ldap_connection_kwargs['password']))

        if seal_and_sign:
            ldap_connection_kwargs['session_security'] = ldap3.ENCRYPT
        elif tls_channel_binding:
            ldap_connection_kwargs['channel_binding'] = ldap3.TLS_CHANNEL_BINDING

        try:
            ldap_connection = ldap3.Connection(**ldap_connection_kwargs)
            ldap_connection.bind()
        except ldap3.core.exceptions.LDAPSocketOpenError as e:
            self._logger.critical(e)
            if self._do_tls:
                self._logger.critical('TLS negociation failed, this error is mostly due to your host '
                                      'not supporting SHA1 as signing algorithm for certificates')
            sys.exit(-1)
        except ldap3.core.exceptions.LDAPInvalidCredentialsResult:
            self._logger.warning('Server returns LDAPInvalidCredentialsResult')
            parsed_error_message = self._parse_ldap_invalid_credentials_result_message(ldap_connection.result['message'])
            if parsed_error_message != "SEC_E_BAD_BINDINGS":
                self._logger.critical('Invalid Credentials : {}'.format(parsed_error_message))
                sys.exit(-1)
            # https://github.com/zyn3rgy/LdapRelayScan#ldaps-channel-binding-token-requirements
            elif parsed_error_message == "SEC_E_BAD_BINDINGS" and not self._tls_channel_binding_supported:
                if self._lmhash and self._nthash:
                    self._logger.critical('Server requires Channel Binding Token and your ldap3 install does not support it. '
                                          'Please run "pip install ldap3-bleeding-edge", try another authentication method, '
                                          'or use a password, not a hash, to authenticate.')
                    sys.exit(-1)
                else:
                    self._logger.debug('Server requires Channel Binding Token but you are using password authentication,'
                                       ' falling back to SIMPLE authentication, hoping LDAPS port is open')
                    ldap_server_kwargs = {'host': self._domain_controller, 'formatter': self._formatter,
                                          'get_info': ldap3.ALL, 'use_ssl': True}
                    server = ldap3.Server(**ldap_server_kwargs)
                    self._do_simple_auth(server)
                    return
            elif self._tls_channel_binding_supported == True and tls_channel_binding == False:
                self._logger.warning('Falling back to TLS with channel binding')
                self._do_ntlm_auth(server, tls_channel_binding=True)
                return

        except ldap3.core.exceptions.LDAPStrongerAuthRequiredResult:
            self._logger.warning('Server returns LDAPStrongerAuthRequiredResult')
            if not self._sign_and_seal_supported:
                self._logger.warning('Sealing not available, falling back to LDAPS')
                # I don't know how to reuse the same Server object, but TLS enabled
                ldap_server_kwargs = {'host': self._domain_controller, 'formatter': self._formatter,
                                      'get_info': ldap3.ALL, 'use_ssl': True}
                server = ldap3.Server(**ldap_server_kwargs)
                self._do_ntlm_auth(server)
                return
            else:
                self._logger.warning('Falling back to NTLM sealing')
                self._do_ntlm_auth(server, seal_and_sign=True)
                return

        who_am_i = ldap_connection.extend.standard.who_am_i()
        self._logger.debug('Successfully connected to the LDAP as {}'.format(who_am_i))
        self._ldap_connection = ldap_connection
        self._server_info = server.info

    def _do_kerberos_auth(self, server, seal_and_sign=False):
        self._logger.debug('LDAP authentication with Kerberos: ldap_scheme = {0} / seal_and_sign = {1}'.format(
                            "ldaps" if server.ssl else "ldap", seal_and_sign))
        ldap_connection_kwargs = {"server": server , 'user': '{}@{}'.format(self._user, self._domain.upper()),
                                  'raise_exceptions': True, 'authentication': ldap3.SASL,
                                  'sasl_mechanism': ldap3.KERBEROS}

        if seal_and_sign:
            ldap_connection_kwargs['session_security'] = ldap3.ENCRYPT

        # Verifying if we have the correct TGS/TGT to interrogate the LDAP server
        ccache = CCache.loadFile(os.getenv('KRB5CCNAME'))
        principal = 'ldap/{}@{}'.format(self._domain_controller.lower(), self._queried_domain.upper())

        # We look for the TGS with the right SPN
        creds = ccache.getCredential(principal, anySPN=False)
        if creds:
            self._logger.debug('TGS found in KRB5CCNAME file')
            creds_server_lower = creds['server'].prettyPrint().lower().decode('utf-8')
            creds_server = creds['server'].prettyPrint().decode('utf-8')
            if creds_server_lower.split('@')[0] != creds_server.split('@')[0]:
                self._logger.debug('SPN not in lowercase, patching SPN')
                new_creds = self._patch_spn(creds, principal)
                # We build a new CCache with the new ticket
                ccache.credentials.append(new_creds)
                temp_ccache = tempfile.NamedTemporaryFile()
                ccache.saveFile(temp_ccache.name)
                cred_store = {'ccache': 'FILE:{}'.format(temp_ccache.name)}
            else:
                self._logger.debug('SPN is good, no patching needed')
                cred_store = dict()
        else:
            self._logger.debug('TGS not found in KRB5CCNAME, looking for '
                    'TGS with alternative SPN')
            # If we don't find it, we search for any SPN
            creds = ccache.getCredential(principal, anySPN=True)
            if creds:
                # If we find one, we build a custom TGS
                self._logger.debug('Alternative TGS found, patching SPN')
                new_creds = self._patch_spn(creds, principal)
                # We build a new CCache with the new ticket
                ccache.credentials.append(new_creds)
                temp_ccache = tempfile.NamedTemporaryFile()
                ccache.saveFile(temp_ccache.name)
                cred_store = {'ccache': 'FILE:{}'.format(temp_ccache.name)}
            else:
                # If we don't find any, we hope for the best (TGT in cache)
                self._logger.debug('Alternative TGS not found, using KRB5CCNAME as is '
                        'while hoping it contains a TGT')
                cred_store = dict()
        ldap_connection_kwargs['cred_store'] = cred_store
        self._logger.debug('LDAP binding parameters: server = {0} / user = {1} '
               '/ Kerberos auth'.format(self._domain_controller, self._user))

        try:
            ldap_connection = ldap3.Connection(**ldap_connection_kwargs)
            ldap_connection.bind()
        except ldap3.core.exceptions.LDAPSocketOpenError as e:
                self._logger.critical(e)
                if self._do_tls:
                    self._logger.critical('TLS negociation failed, this error is mostly due to your host '
                                          'not supporting SHA1 as signing algorithm for certificates')
                sys.exit(-1)

        except ldap3.core.exceptions.LDAPStrongerAuthRequiredResult:
            self._logger.warning('Server returns LDAPStrongerAuthRequiredResult')
            if not self._sign_and_seal_supported:
                self._logger.warning('Sealing not available, falling back to LDAPS')
                server.use_ssl = True
                # I don't know how to reuse the same Server object, but switched to TLS
                ldap_server_kwargs = {'host': self._domain_controller, 'formatter': self._formatter,
                                      'get_info': ldap3.ALL, 'use_ssl': True}
                server = ldap3.Server(**ldap_server_kwargs)
                self._do_kerberos_auth(server)
                return
            else:
                self._logger.warning('Falling back to Kerberos sealing')
                self._do_kerberos_auth(server, seal_and_sign=True)
                return

        who_am_i = ldap_connection.extend.standard.who_am_i()
        self._logger.debug('Successfully connected to the LDAP as {}'.format(who_am_i))
        self._ldap_connection = ldap_connection
        self._server_info = server.info

    def _do_schannel_auth(self, server):
        self._logger.debug('LDAP authentication with SChannel: ldap_scheme = {0}'.format("ldaps" if server.ssl else "ldap"))
        tls_mode = 'Implicit'
        server.tls = ldap3.Tls(local_private_key_file=self._user_key, local_certificate_file=self._user_cert, validate=ssl.CERT_NONE)
        ldap_connection_kwargs = {'server': server, 'user': None, 'raise_exceptions': True}

        # Explicit TLS, setting up StartTLS
        if not self._do_tls:
            tls_mode = 'Explicit'
            self._logger.warning('Using certificate authentication but --tls not provided, setting up TLS with StartTLS')
            ldap_connection_kwargs['authentication'] = ldap3.SASL
            ldap_connection_kwargs['sasl_mechanism'] = ldap3.EXTERNAL

        self._logger.debug('LDAP binding parameters: server = {0} / cert = {1} '
            '/ key = {2} / {3} TLS / SChannel auth'.format(self._domain_controller, self._user_cert, self._user_key, tls_mode))

        try:
            ldap_connection = ldap3.Connection(**ldap_connection_kwargs)
            ldap_connection.open()
            ldap_connection.raise_exceptions = False
            if not self._do_tls:
                self._logger.debug('Sending StartTLS command')
                if ldap_connection.start_tls():
                    self._logger.debug('StartTLS succeeded')
                    ldap_connection.raise_exceptions = True
                    ldap_connection.bind()
                else:
                    self._logger.critical('StartTLS failed, exiting')
                    sys.exit(-1)
        except Exception as e:
            # I don't really understand exception when using SChannel authentication, but if you see this message
            # your cert and key are probably not valid
            self._logger.critical('Exception during SChannel authentication: {}'.format(e))
            sys.exit(-1)

        who_am_i = ldap_connection.extend.standard.who_am_i()

        if not who_am_i:
            self._logger.critical('Certificate authentication failed')
            sys.exit(-1)

        self._logger.debug('Successfully connected to the LDAP as {}'.format(who_am_i))
        self._ldap_connection = ldap_connection
        self._server_info = server.info

    def _do_simple_auth(self, server):
        self._logger.debug('LDAP authentication with SIMPLE: ldap_scheme = {0}'.format("ldaps" if server.ssl else "ldap"))
        if self._password is None:
            self._logger.critical('SIMPLE authentication only supports password')
            sys.exit(-1)
        ldap_connection_kwargs = {'server': server, 'user': '{}@{}'.format(self._user, self._domain),
                                  'password': self._password, 'raise_exceptions': True, 'authentication': ldap3.SIMPLE}

        self._logger.debug('LDAP binding parameters: server = {0} / user = {1} '
            '/ password = {2}'.format(self._domain_controller, self._user, self._password))

        try:
            ldap_connection = ldap3.Connection(**ldap_connection_kwargs)
            ldap_connection.bind()
        except ldap3.core.exceptions.LDAPSocketOpenError as e:
                self._logger.critical(e)
                if self._do_tls:
                    self._logger.critical('TLS negociation failed, this error is mostly due to your host '
                                          'not supporting SHA1 as signing algorithm for certificates')
                sys.exit(-1)
        except ldap3.core.exceptions.LDAPInvalidCredentialsResult as e:
            parsed_error_message = self._parse_ldap_invalid_credentials_result_message(ldap_connection.result['message'])
            self._logger.critical('Invalid Credentials : {}'.format(parsed_error_message))
            sys.exit(-1)

        who_am_i = ldap_connection.extend.standard.who_am_i()
        self._logger.debug('Successfully connected to the LDAP as {}'.format(who_am_i))
        self._ldap_connection = ldap_connection
        self._server_info = server.info

    def _get_netfqdn(self):
        server = ldap3.Server(self._domain_controller, get_info=ldap3.ALL)
        connection = ldap3.Connection(server)
        try:
            connection.bind()
            self._logger.debug('retrieved defaultnamingcontext: {}'.format(server.info.other["defaultNamingContext"][0]))
            fqdn = server.info.other["defaultNamingContext"][0][3:].replace(',DC=','.')
        except:
            self._logger.warning('Error when binding anonymously to the LDAP. Retry with -d and -w options.')
            sys.exit(-1)
        self._logger.debug('returned fqdn: {}'.format(fqdn))
        return fqdn

    def _patch_spn(self, creds, principal):
        self._logger.debug('Patching principal to {}'.format(principal))

        from pyasn1.codec.der import decoder, encoder
        from impacket.krb5.asn1 import TGS_REP, Ticket

        # Code is ~~based on~~ stolen from https://github.com/SecureAuthCorp/impacket/pull/1256
        tgs = creds.toTGS(principal)
        decoded_st = decoder.decode(tgs['KDC_REP'], asn1Spec=TGS_REP())[0]
        decoded_st['ticket']['sname']['name-string'][0] = 'ldap'
        decoded_st['ticket']['sname']['name-string'][1] = self._domain_controller.lower()
        decoded_st['ticket']['realm'] = self._queried_domain.upper()

        new_creds = Credential(data=creds.getData())
        new_creds.ticket = CountedOctetString()
        new_creds.ticket['data'] = encoder.encode(decoded_st['ticket'].clone(tagSet=Ticket.tagSet, cloneValueFlag=True))
        new_creds.ticket['length'] = len(new_creds.ticket['data'])
        new_creds['server'].fromPrincipal(Principal(principal, type=constants.PrincipalNameType.NT_PRINCIPAL.value))

        return new_creds

    def _create_ldap_connection(self, queried_domain=str(), ads_path=str(),
                                ads_prefix=str()):
        if not queried_domain:
            self._logger.warning('Queried domain not specified, assuming it is the same than the target')
            queried_domain = self._get_netfqdn()

        self._queried_domain = queried_domain

        if not self._domain:
            if self._do_certificate:
                self._logger.critical('-w with the FQDN must be used when authenticating with certificates')
                sys.exit(-1)
            elif self._do_kerberos:
                ccache = CCache.loadFile(os.getenv('KRB5CCNAME'))
                self._domain = ccache.principal.realm['data'].decode('utf-8')
            else:
                self._logger.warning('User domain not specified, assuming it is the same than the queried domain')
                self._domain = queried_domain

        ldap_server_kwargs = {'host': self._domain_controller, 'formatter': self._formatter,
                              'get_info': ldap3.ALL}

        if self._do_tls:
            ldap_server_kwargs['use_ssl'] = True
            self._logger.debug('LDAPS connection forced')
        else:
            ldap_server_kwargs['use_ssl'] = False

        ldap_server = ldap3.Server(**ldap_server_kwargs)

        if self._do_kerberos:
            self._do_kerberos_auth(ldap_server)
        elif self._do_certificate:
            self._do_schannel_auth(ldap_server)
        elif self._do_simple:
            self._do_simple_auth(ldap_server)
        else:
            self._do_ntlm_auth(ldap_server)

        base_dn = str()

        if ads_prefix:
            self._ads_prefix = ads_prefix
            base_dn = '{},'.format(self._ads_prefix)
        if ads_path:
            # TODO: manage ADS path starting with 'GC://'
            if ads_path.upper().startswith('LDAP://'):
                ads_path = ads_path[7:]
            self._ads_path = ads_path
            base_dn += self._ads_path
        else:
            base_dn += self._server_info.other["defaultNamingContext"][0]

        # base_dn is no longer used within `_create_ldap_connection()`, but I don't want to break
        # the function call. So we store it in an attriute and use it in `_ldap_search()`
        self._base_dn = base_dn

    def _ldap_search(self, search_filter, class_result, attributes=list(), controls=list()):
        results = list()

        # if no attribute name specified, we return all attributes
        if not attributes:
            attributes =  ldap3.ALL_ATTRIBUTES

        self._logger.debug('search_base = {0} / search_filter = {1} / attributes = {2}'.format(self._base_dn,
                                                                                               search_filter,
                                                                                               attributes))

        # Microsoft Active Directory set an hard limit of 1000 entries returned by any search
        search_results=self._ldap_connection.extend.standard.paged_search(search_base=self._base_dn,
                search_filter=search_filter, attributes=attributes,
                controls=controls, paged_size=1000, generator=True)

        try:
            # Skip searchResRef
            for result in search_results:
                if result['type'] != 'searchResEntry':
                    continue
                results.append(class_result(result['attributes']))

        except ldap3.core.exceptions.LDAPAttributeError as e:
            self._logger.critical(e)
            sys.exit(-1)

        if not results:
            self._logger.debug('Query returned an empty result')

        return results

    @staticmethod
    def _ldap_connection_init(f):
        def wrapper(*args, **kwargs):
            instance = args[0]
            queried_domain = kwargs.get('queried_domain', None)
            ads_path = kwargs.get('ads_path', None)
            ads_prefix = kwargs.get('ads_prefix', None)
            if (not instance._ldap_connection) or \
               (queried_domain and queried_domain != instance._queried_domain) or \
               (ads_path and ads_path != instance._ads_path) or \
               (ads_prefix and ads_prefix != instance._ads_prefix):
                if instance._ldap_connection:
                    instance._ldap_connection.unbind()
                instance._create_ldap_connection(queried_domain=queried_domain,
                                                 ads_path=ads_path, ads_prefix=ads_prefix)
            return f(*args, **kwargs)
        return wrapper

    def __enter__(self):
        self._create_ldap_connection()
        return self

    def __exit__(self, type, value, traceback):
        try:
            self._ldap_connection.unbind()
        except AttributeError:
            self._logger.warning('Error when unbinding')
            pass
        self._ldap_connection = None

class RPCRequester():
    def __init__(self, target_computer, domain=str(), user=(), password=str(),
                 lmhash=str(), nthash=str(), do_kerberos=False):
        self._target_computer = target_computer
        self._domain = domain
        self._user = user
        self._password = password
        self._lmhash = lmhash
        self._nthash = nthash
        self._do_kerberos = do_kerberos
        self._pipe = None
        self._rpc_connection = None
        self._dcom = None
        self._wmi_connection = None

        logger = logging.getLogger('pywerview_main_logger.RPCRequester')
        logger.ULTRA = 5
        self._logger = logger

    def _create_rpc_connection(self, pipe):
        # Here we build the DCE/RPC connection
        self._pipe = pipe

        binding_strings = dict()
        binding_strings['srvsvc'] = srvs.MSRPC_UUID_SRVS
        binding_strings['wkssvc'] = wkst.MSRPC_UUID_WKST
        binding_strings['samr'] = samr.MSRPC_UUID_SAMR
        binding_strings['svcctl'] = scmr.MSRPC_UUID_SCMR
        binding_strings['drsuapi'] = drsuapi.MSRPC_UUID_DRSUAPI

        # TODO: try to fallback to TCP/139 if tcp/445 is closed
        if self._pipe == r'\drsuapi':
            string_binding = epm.hept_map(self._target_computer, drsuapi.MSRPC_UUID_DRSUAPI,
                                          protocol='ncacn_ip_tcp')
            rpctransport = transport.DCERPCTransportFactory(string_binding)
            rpctransport.set_credentials(username=self._user, password=self._password,
                                         domain=self._domain, lmhash=self._lmhash,
                                         nthash=self._nthash)
        else:
            rpctransport = transport.SMBTransport(self._target_computer, 445, self._pipe,
                                                  username=self._user, password=self._password,
                                                  domain=self._domain, lmhash=self._lmhash,
                                                  nthash=self._nthash, doKerberos=self._do_kerberos)

        rpctransport.set_connect_timeout(10)
        dce = rpctransport.get_dce_rpc()

        if self._pipe == r'\drsuapi':
            dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)

        try:
            dce.connect()
        except Exception as e:
            self._logger.warning('Error when creating RPC connection')
            self._logger.warning(e)
            self._rpc_connection = None
        else:
            dce.bind(binding_strings[self._pipe[1:]])
            self._rpc_connection = dce

    def _create_wmi_connection(self, namespace='root\\cimv2', rpc_auth_level=None):
        self._namespace = namespace
        self._rpc_auth_level = rpc_auth_level

        try:
            self._dcom = DCOMConnection(self._target_computer, self._user, self._password,
                                        self._domain, self._lmhash, self._nthash, doKerberos=self._do_kerberos)
        except Exception as e:
            self._logger.warning('Error when creating WMI connection')
            self._logger.warning(e)
            self._dcom = None
        else:
            i_interface = self._dcom.CoCreateInstanceEx(wmi.CLSID_WbemLevel1Login,
                                                        wmi.IID_IWbemLevel1Login)
            i_wbem_level1_login = wmi.IWbemLevel1Login(i_interface)
            self._wmi_connection = i_wbem_level1_login.NTLMLogin(ntpath.join('\\\\{}\\'.format(self._target_computer), self._namespace),
                                                                 NULL, NULL)
            if self._rpc_auth_level == 'privacy':
                self._wmi_connection.get_dce_rpc().set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)
            elif self._rpc_auth_level == 'integrity':
                self._wmi_connection.get_dce_rpc().set_auth_level(RPC_C_AUTHN_LEVEL_PKT_INTEGRITY)

    @staticmethod
    def _rpc_connection_init(pipe=r'\srvsvc'):
        def decorator(f):
            def wrapper(*args, **kwargs):
                instance = args[0]
                if (not instance._rpc_connection) or (pipe != instance._pipe):
                    if instance._rpc_connection:
                        instance._rpc_connection.disconnect()
                    instance._create_rpc_connection(pipe=pipe)
                if instance._rpc_connection is None:
                    return None
                return f(*args, **kwargs)
            return wrapper
        return decorator

    @staticmethod
    def _wmi_connection_init(namespace='root\\cimv2', rpc_auth_level=None):
        def decorator(f):
            def wrapper(*args, **kwargs):
                instance = args[0]
                if (not instance._wmi_connection) or (namespace != instance._namespace) \
                        or (rpc_auth_level != instance._rpc_auth_level):
                    instance._create_wmi_connection()
                if instance._dcom is None:
                    return None
                return f(*args, **kwargs)
            return wrapper
        return decorator

    def __enter__(self):
        # Picked because it's the most used by the net* functions
        self._create_rpc_connection(r'\srvsvc')
        return self

    def __exit__(self, type, value, traceback):
        try:
            self._rpc_connection.disconnect()
        except AttributeError:
            pass
        self._rpc_connection = None

class LDAPRPCRequester(LDAPRequester, RPCRequester):
    def __init__(self, target_computer, domain=str(), user=str(), password=str(),
                 lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False, do_tls=False,
                 user_cert=str(), user_key=str(), domain_controller=str()):
        # If no domain controller was given, we assume that the user wants to
        # target a domain controller to perform LDAP requests against
        if not domain_controller:
            domain_controller = target_computer

        LDAPRequester.__init__(self, domain_controller, domain, user, password,
                               lmhash, nthash, do_simple, do_kerberos, do_tls,
                               user_cert, user_key)

        if user_cert is not None and user_key is not None and do_simple is not None:
            RPCRequester.__init__(self, target_computer, domain, user, password,
                                  lmhash, nthash, do_kerberos)

        logger = logging.getLogger('pywerview_main_logger.LDAPRPCRequester')
        logger.ULTRA = 5
        self._logger = logger

    def __enter__(self):
        try:
            LDAPRequester.__enter__(self)
        except (socket.error, IndexError):
            pass

        try:
            RPCRequester.__enter__(self)
        except (AttributeError):
            pass

        return self

    def __exit__(self, type, value, traceback):
        LDAPRequester.__exit__(self, type, value, traceback)
        RPCRequester.__exit__(self, type, value, traceback)




# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

import logging
from multiprocessing import Process, Pipe

from pywerview.functions.net import NetRequester
from pywerview.functions.misc import Misc
import pywerview.objects.rpcobjects as rpcobj
from impacket.dcerpc.v5.rpcrt import DCERPCException

class HunterWorker(Process):
    def __init__(self, pipe, domain, user, password, lmhash, nthash, do_kerberos, do_tls):
        Process.__init__(self)
        self._pipe = pipe
        self._domain = domain
        self._user = user
        self._password = password
        self._lmhash = lmhash
        self._nthash = nthash
        self._do_kerberos = do_kerberos
        self._do_tls = do_tls

        logger = logging.getLogger('pywerview_main_logger.HunterWorker')
        logger.ULTRA = 5
        self._logger = logger

    def terminate(self):
        self._pipe.close()
        Process.terminate(self)

    def run(self):
        while True:
            target_computer = self._pipe.recv()
            result = self._hunt(target_computer)
            self._pipe.send(result)

class UserHunterWorker(HunterWorker):
    def __init__(self, pipe, domain, user, password, lmhash, nthash, do_kerberos, do_tls,
            foreign_users, stealth, target_users, domain_short_name, check_access):
        HunterWorker.__init__(self, pipe, domain, user, password, lmhash,
                nthash, do_kerberos, do_tls)
        self._foreign_users = foreign_users
        self._stealth = stealth
        self._target_users = target_users
        self._domain_short_name = domain_short_name
        self._check_access = check_access

    # TODO: test foreign user hunting
    def _hunt(self, target_computer):
        # TODO: implement ping of target
        results = list()
        # First, we get every distant session on the target computer
        distant_sessions = list()
        self._logger.debug('Start hunting user on {}'.format(target_computer))
        net_requester =  NetRequester(target_computer, self._domain, self._user, self._password,
                          self._lmhash, self._nthash, self._do_kerberos, self._do_tls)
        try:
            if not self._foreign_users:
                self._logger.log(self._logger.ULTRA, 'Calling get_netsession on {}'.format(target_computer))
                distant_sessions += net_requester.get_netsession()
            if not self._stealth:
                self._logger.log(self._logger.ULTRA, 'Calling get_netloggedon on {}'.format(target_computer))
                distant_sessions += net_requester.get_netloggedon()
        except TypeError:
            self._logger.warning('Error when retrieving sessions, skipping {}...'.format(target_computer))
            return results


        self._logger.debug('{} distant sessions found'.format(len(distant_sessions)))
        self._logger.log(self._logger.ULTRA,'Distant sessions: {}'.format(distant_sessions))

        # For every session, we get information on the remote user
        for session in distant_sessions:
            try:
                username = session.sesi10_username
                userdomain = str()
                session_from = session.sesi10_cname
                if session_from.startswith('\\'):
                    session_from = session_from.lstrip('\\')
            except AttributeError:
                username = session.wkui1_username
                userdomain = session.wkui1_logon_domain
                session_from = str()

            # If we found a user
            if username:
                self._logger.log(self._logger.ULTRA, 'User found in session ({})'.format(username))
                # We see if it's in our target user group
                for target_user in self._target_users:
                    if target_user.membername.lower() in username.lower():

                        self._logger.log(self._logger.ULTRA, 'We found our target! ({})'.format(username))

                        # If we fall in this branch, we're looking for foreign users
                        # and found a user in the same domain
                        if self._domain_short_name and self._domain_short_name.lower() == userdomain.lower():
                            continue

                        attributes = dict()
                        if userdomain:
                            attributes['userdomain'] = userdomain
                        else:
                            attributes['userdomain'] = target_user.memberdomain
                        attributes['username'] = username
                        attributes['computername'] = target_computer
                        attributes['sessionfrom'] = session_from

                        if self._check_access:
                            self._logger.debug('"Check access" requested, calling invoke-checklocaladminaccess')
                            with Misc(target_computer, self._domain, self._user, self._password,
                                              self._lmhash, self._nthash, self._do_kerberos) as misc_requester:
                                attributes['localadmin'] = misc_requester.invoke_checklocaladminaccess()
                        else:
                            attributes['localadmin'] = str()

                        results.append(rpcobj.RPCObject(attributes))

        self._logger.debug('Target\' sessions found on {0}: {1}'.format(target_computer, len(results)))
        return results

class ProcessHunterWorker(HunterWorker):
    def __init__(self, pipe, domain, user, password, lmhash, nthash, do_kerberos, do_tls,
            process_name, target_users):
        HunterWorker.__init__(self, pipe, domain, user, password, lmhash, nthash, do_kerberos, do_tls)
        self._process_name = process_name
        self._target_users = target_users

    def _hunt(self, target_computer):
        results = list()

        self._logger.debug('Start hunting process on {}'.format(target_computer))

        distant_processes = list()
        net_requester = NetRequester(target_computer, self._domain, self._user, self._password,
                                     self._lmhash, self._nthash, self._do_kerberos, self._do_tls)
        
        self._logger.log(self._logger.ULTRA, 'Calling get_netprocess on {}'.format(target_computer))

        try:
            distant_processes = net_requester.get_netprocess()
        except DCERPCException:
            self._logger.warning('Error when retrieving process, skipping {}...'.format(target_computer))
            return results

        try:
            for process in distant_processes:
                if self._process_name:
                    for process_name in self._process_name:
                        if process_name.lower() in process.processname.lower():
                            self._logger.log(self._logger.ULTRA,'Found processname {0} on {1}'.format(process_name, target_computer))
                            results.append(process)
                elif self._target_users:
                    for target_user in self._target_users:
                        if target_user.membername.lower() in process.user.lower():
                            self._logger.log(self._logger.ULTRA,'Found {0} process on {1}'.format(target_user.membername, target_computer))
                        results.append(process)
        except TypeError:
            self._logger.warning('Error when handling process, skipping {}...'.format(target_computer))
            return results

        self._logger.debug('Processname found on {0}: {1}'.format(target_computer, len(results)))
        return results

class EventHunterWorker(HunterWorker):
    def __init__(self, pipe, domain, user, password, lmhash, nthash, do_kerberos, do_tls,
            search_days, target_users):
        HunterWorker.__init__(self, pipe, domain, user, password, lmhash, nthash, do_kerberos, do_tls)
        self._target_users = target_users
        self._search_days = search_days

    def _hunt(self, target_computer):
        results = list()

        self._logger.debug('Start hunting event on {}'.format(target_computer))

        distant_processes = list()
        net_requester = NetRequester(target_computer, self._domain, self._user, self._password,
                                     self._lmhash, self._nthash, self._do_kerberos, self._do_tls)
        
        self._logger.log(self._logger.ULTRA, 'Calling get_userevent on {}'.format(target_computer))
        distant_events = net_requester.get_userevent(date_start=self._search_days)

        try:
            for event in distant_events:
                if self._target_users:
                    for target_user in self._target_users:
                        if target_user.membername.lower() in event.username.lower():
                            results.append(event)
        except TypeError:
            self._logger.warning('Error when retrieving event, skipping {}...'.format(target_computer))
        return results

